@@ -53,18 +53,67 @@ typedef ssize_t Py_ssize_t;
 #define PYLIBMC_SERVER_UDP   (1 << 1)
 #define PYLIBMC_SERVER_UNIX  (1 << 2)
 
-/* Key flags from python-memcache. */
+/* {{{ Key flags from python-memcached
+ * Some flags (like the compression one, ZLIB) are combined with others.
+ */
 #define PYLIBMC_FLAG_NONE    0
 #define PYLIBMC_FLAG_PICKLE  (1 << 0)
 #define PYLIBMC_FLAG_INTEGER (1 << 1)
 #define PYLIBMC_FLAG_LONG    (1 << 2)
-#define PYLIBMC_FLAG_BOOL    (1 << 3)
-
-#define PYLIBMC_INC  (1 << 0)
-#define PYLIBMC_DEC  (1 << 1)
+/* Note: this is an addition! python-memcached doesn't handle bools. */
+#define PYLIBMC_FLAG_BOOL    (1 << 4)
+#define PYLIBMC_FLAG_TYPES   (PYLIBMC_FLAG_PICKLE | PYLIBMC_FLAG_INTEGER | \
+                              PYLIBMC_FLAG_LONG | PYLIBMC_FLAG_BOOL)
+/* Modifier flags */
+#define PYLIBMC_FLAG_ZLIB    (1 << 3)
+/* }}} */
 
 typedef memcached_return (*_PylibMC_SetCommand)(memcached_st *, const char *,
         size_t, const char *, size_t, time_t, uint32_t);
+typedef memcached_return (*_PylibMC_IncrCommand)(memcached_st *,
+        const char *, size_t, unsigned int, uint64_t*);
+
+typedef struct {
+  char key[MEMCACHED_MAX_KEY];
+  size_t key_len;
+  char *value;
+  size_t value_len;
+  uint32_t flags;
+} pylibmc_mget_result;
+
+typedef struct {
+  char *key;
+  Py_ssize_t key_len;
+  char* value;
+  Py_ssize_t value_len;
+  time_t time;
+  uint32_t flags;
+
+  /* the objects that must be freed after the mset is executed */
+  PyObject* key_obj;
+  PyObject* prefixed_key_obj;
+  PyObject* value_obj;
+
+  /* the success of executing the mset afterwards */
+  int success;
+
+} pylibmc_mset;
+
+typedef struct {
+  char* key;
+  Py_ssize_t key_len;
+  _PylibMC_IncrCommand incr_func;
+  unsigned int delta;
+  uint64_t result;
+} pylibmc_incr;
+
+typedef struct {
+    PyObject *self;
+    PyObject *retval;
+    memcached_server_st *servers;  /* DEPRECATED */
+    memcached_stat_st *stats;
+    int index;
+} _PylibMC_StatsContext;
 
 /* {{{ Exceptions */
 static PyObject *PylibMCExc_MemcachedError;
@@ -124,26 +173,21 @@ typedef struct {
 } PylibMC_Behavior;
 
 static PylibMC_Behavior PylibMC_behaviors[] = {
-    { MEMCACHED_BEHAVIOR_NO_BLOCK, "no block" },
+    { MEMCACHED_BEHAVIOR_NO_BLOCK, "no_block" },
     { MEMCACHED_BEHAVIOR_TCP_NODELAY, "tcp_nodelay" },
     { MEMCACHED_BEHAVIOR_HASH, "hash" },
-    { MEMCACHED_BEHAVIOR_KETAMA, "ketama" },
-    { MEMCACHED_BEHAVIOR_DISTRIBUTION, "distribution" },
-    { MEMCACHED_BEHAVIOR_SOCKET_SEND_SIZE, "socket send size" },
-    { MEMCACHED_BEHAVIOR_SOCKET_RECV_SIZE, "socket recv size" },
-    { MEMCACHED_BEHAVIOR_CACHE_LOOKUPS, "cache_lookups" },
-    { MEMCACHED_BEHAVIOR_SUPPORT_CAS, "support_cas" },
-    { MEMCACHED_BEHAVIOR_POLL_TIMEOUT, "poll_timeout" },
-    { MEMCACHED_BEHAVIOR_BUFFER_REQUESTS, "buffer_requests" },
-    { MEMCACHED_BEHAVIOR_SORT_HOSTS, "sort_hosts" },
-    { MEMCACHED_BEHAVIOR_VERIFY_KEY, "verify_key" },
-    { MEMCACHED_BEHAVIOR_CONNECT_TIMEOUT, "connect_timeout" },
-    { MEMCACHED_BEHAVIOR_RETRY_TIMEOUT, "retry_timeout" },
-    { MEMCACHED_BEHAVIOR_KETAMA_WEIGHTED, "ketama_weighted" },
     { MEMCACHED_BEHAVIOR_KETAMA_HASH, "ketama_hash" },
-    { MEMCACHED_BEHAVIOR_SND_TIMEOUT, "snd_timeout" },
-    { MEMCACHED_BEHAVIOR_RCV_TIMEOUT, "rcv_timeout" },
-    { MEMCACHED_BEHAVIOR_SERVER_FAILURE_LIMIT, "server_failure_limit" },
+    { MEMCACHED_BEHAVIOR_KETAMA, "ketama" },
+    { MEMCACHED_BEHAVIOR_KETAMA_WEIGHTED, "ketama_weighted" },
+    { MEMCACHED_BEHAVIOR_DISTRIBUTION, "distribution" },
+    { MEMCACHED_BEHAVIOR_CACHE_LOOKUPS, "cache_lookups" },
+    { MEMCACHED_BEHAVIOR_SUPPORT_CAS, "cas" },
+    { MEMCACHED_BEHAVIOR_BUFFER_REQUESTS, "buffer_requests" },
+    { MEMCACHED_BEHAVIOR_VERIFY_KEY, "verify_keys" },
+    { MEMCACHED_BEHAVIOR_CONNECT_TIMEOUT, "connect_timeout" },
+    { MEMCACHED_BEHAVIOR_SND_TIMEOUT, "send_timeout" },
+    { MEMCACHED_BEHAVIOR_RCV_TIMEOUT, "receive_timeout" },
+    { MEMCACHED_BEHAVIOR_SERVER_FAILURE_LIMIT, "failure_limit" },
     { MEMCACHED_BEHAVIOR_IO_MSG_WATERMARK, "_io_msg_watermark" },
     { MEMCACHED_BEHAVIOR_IO_BYTES_WATERMARK, "_io_bytes_watermark" },
     { MEMCACHED_BEHAVIOR_IO_KEY_PREFETCH, "_io_key_prefetch" },
@@ -151,6 +195,11 @@ static PylibMC_Behavior PylibMC_behaviors[] = {
     { MEMCACHED_BEHAVIOR_NOREPLY, "_noreply" },
     { MEMCACHED_BEHAVIOR_AUTO_EJECT_HOSTS, "_auto_eject_hosts" },
     { MEMCACHED_BEHAVIOR_NUMBER_OF_REPLICAS, "_number_of_replicas" },
+    { MEMCACHED_BEHAVIOR_SORT_HOSTS, "_sort_hosts" },
+    { MEMCACHED_BEHAVIOR_RETRY_TIMEOUT, "_retry_timeout" },
+    { MEMCACHED_BEHAVIOR_POLL_TIMEOUT, "_poll_timeout" },
+    { MEMCACHED_BEHAVIOR_SOCKET_SEND_SIZE, "_socket_send_size" },
+    { MEMCACHED_BEHAVIOR_SOCKET_RECV_SIZE, "_socket_recv_size" },
     { 0, NULL }
 };
 
@@ -162,8 +211,10 @@ static PylibMC_Behavior PylibMC_hashers[] = {
     { MEMCACHED_HASH_FNV1A_64, "fnv1a_64" },
     { MEMCACHED_HASH_FNV1_32, "fnv1_32" },
     { MEMCACHED_HASH_FNV1A_32, "fnv1a_32" },
-    { MEMCACHED_HASH_HSIEH, "hsieh" },
     { MEMCACHED_HASH_MURMUR, "murmur" },
+#ifdef MEMCACHED_HASH_HSIEH
+    { MEMCACHED_HASH_HSIEH, "hsieh" },
+#endif
     { 0, NULL }
 };
 
@@ -195,8 +246,10 @@ static PyObject *PylibMC_Client_append(PylibMC_Client *, PyObject *, PyObject *)
 static PyObject *PylibMC_Client_delete(PylibMC_Client *, PyObject *);
 static PyObject *PylibMC_Client_incr(PylibMC_Client *, PyObject *);
 static PyObject *PylibMC_Client_decr(PylibMC_Client *, PyObject *);
+static PyObject *PylibMC_Client_incr_multi(PylibMC_Client*, PyObject*, PyObject*);
 static PyObject *PylibMC_Client_get_multi(PylibMC_Client *, PyObject *, PyObject *);
 static PyObject *PylibMC_Client_set_multi(PylibMC_Client *, PyObject *, PyObject *);
+static PyObject *PylibMC_Client_add_multi(PylibMC_Client *, PyObject *, PyObject *);
 static PyObject *PylibMC_Client_delete_multi(PylibMC_Client *, PyObject *, PyObject *);
 static PyObject *PylibMC_Client_get_behaviors(PylibMC_Client *);
 static PyObject *PylibMC_Client_set_behaviors(PylibMC_Client *, PyObject *);
@@ -210,6 +263,24 @@ static PyObject *_PylibMC_Unpickle(const char *, size_t);
 static PyObject *_PylibMC_Pickle(PyObject *);
 static int _PylibMC_CheckKey(PyObject *);
 static int _PylibMC_CheckKeyStringAndSize(char *, Py_ssize_t);
+static int _PylibMC_SerializeValue(PyObject* key_obj,
+                                   PyObject* key_prefix,
+                                   PyObject* value_obj,
+                                   time_t time,
+                                   pylibmc_mset* serialized);
+static void _PylibMC_FreeMset(pylibmc_mset*);
+static PyObject *_PylibMC_RunSetCommandSingle(PylibMC_Client *self,
+        _PylibMC_SetCommand f, char *fname, PyObject *args, PyObject *kwds);
+static PyObject *_PylibMC_RunSetCommandMulti(PylibMC_Client* self,
+        _PylibMC_SetCommand f, char *fname, PyObject *args, PyObject *kwds);
+static bool _PylibMC_RunSetCommand(PylibMC_Client* self,
+                                   _PylibMC_SetCommand f, char *fname,
+                                   pylibmc_mset* msets, size_t nkeys,
+                                   size_t min_compress);
+static int _PylibMC_Deflate(char* value, size_t value_len,
+                            char** result, size_t *result_len);
+static bool _PylibMC_IncrDecr(PylibMC_Client*, pylibmc_incr*, size_t);
+
 /* }}} */
 
 /* {{{ Type's method table */
@@ -232,10 +303,14 @@ static PyMethodDef PylibMC_ClientType_methods[] = {
         "Increment a key by a delta."},
     {"decr", (PyCFunction)PylibMC_Client_decr, METH_VARARGS,
         "Decrement a key by a delta."},
+    {"incr_multi", (PyCFunction)PylibMC_Client_incr_multi, METH_VARARGS|METH_KEYWORDS,
+        "Increment more than one key by a delta."},
     {"get_multi", (PyCFunction)PylibMC_Client_get_multi,
         METH_VARARGS|METH_KEYWORDS, "Get multiple keys at once."},
     {"set_multi", (PyCFunction)PylibMC_Client_set_multi,
         METH_VARARGS|METH_KEYWORDS, "Set multiple keys at once."},
+    {"add_multi", (PyCFunction)PylibMC_Client_add_multi,
+        METH_VARARGS|METH_KEYWORDS, "Add multiple keys at once."},
     {"delete_multi", (PyCFunction)PylibMC_Client_delete_multi,
         METH_VARARGS|METH_KEYWORDS, "Delete multiple keys at once."},
     {"get_behaviors", (PyCFunction)PylibMC_Client_get_behaviors, METH_NOARGS,
@@ -289,23 +364,23 @@ static PyTypeObject PylibMC_ClientType = {
     0,
     0,
     PylibMC_ClientType_methods,
-    0,                                   
-    0,                                   
-    0,                                   
-    0,                                   
-    0,                                   
-    0,                                   
-    0,                                   
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
     (initproc)PylibMC_Client_init,
-    0,                                   
-    (newfunc)PylibMC_ClientType_new,              
-    0,          
-    0,          
-    0,          
-    0,          
-    0,          
-    0,          
-    0,          
+    0,
+    (newfunc)PylibMC_ClientType_new, //PyType_GenericNew,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
     0
 };
 
