@@ -45,6 +45,16 @@ Batch operation::
     {'cats': ['on acid', 'furry'], 'dogs': True}
     >>> mc.delete_multi(["cats", "dogs", "nonextant"])
     False
+    >>> mc.add_multi({"cats": ["on acid", "furry"], "dogs": True})
+    []
+    >>> mc.get_multi(["cats", "dogs"])
+    {'cats': ['on acid', 'furry'], 'dogs': True}
+    >>> mc.add_multi({"cats": "not set", "dogs": "definitely not set", "bacon": "yummy"})
+    ['cats', 'dogs']
+    >>> mc.get_multi(["cats", "dogs", "bacon"])
+    {'cats': ['on acid', 'furry'], 'bacon': 'yummy', 'dogs': True}
+    >>> mc.delete_multi(["cats", "dogs", "bacon"])
+    True
 
 Further Reading
 ===============
@@ -53,7 +63,6 @@ See http://sendapatch.se/projects/pylibmc/
 """
 
 import _pylibmc
-from warnings import warn
 from Queue import Queue
 
 __all__ = ["hashers", "distributions", "Client"]
@@ -148,11 +157,6 @@ class Client(_pylibmc.client):
         This also happens for `distribution`.
         """
         behaviors = behaviors.copy()
-        if any(" " in k for k in behaviors):
-            warn(DeprecationWarning("space-delimited behavior names "
-                                    "are deprecated"))
-            for k in [k for k in behaviors if " " in k]:
-                behaviors[k.replace(" ", "_")] = behaviors.pop(k)
         if behaviors.get("hash") is not None:
             behaviors["hash"] = hashers[behaviors["hash"]]
         if behaviors.get("ketama_hash") is not None:
@@ -260,5 +264,20 @@ except ImportError:
     del ThreadMappedPool
 
 if __name__ == "__main__":
-    import doctest
-    doctest.testmod()
+    import sys
+    import code
+
+    svr_addrs = []
+
+    sys.stderr.write("pylibmc interactive shell\n\n")
+    sys.stderr.write("Input list of servers, end by a blank line\n")
+
+    while True:
+        in_addr = raw_input("Address: ")
+        if not in_addr:
+            break
+    if not svr_addrs:
+        svr_addrs.append("127.0.0.1")
+
+    mc = Client(svr_addrs)
+    code.interact(banner="\nmc client available as `mc`", local={"mc": mc})
